#include <libgen.h>
#include <string.h>

/* XXX: might want to rename 'statetransferd.h', because the client needs it, too.. */
#include "statetransferd.h"

#define is_flag(f) (!strcmp(f, flag))
#define usage() __usage(basename(argv[0]))

static CLIENT *clnt;


void __usage(const char *name)
{
	const char *oosage = \
	"%s -- RPC client program for NFSv4 statetransfer.\n"
	"usage:  %s -l|-p|-g|-x|-t <server> | -a|-r <recoverydir> <client> <server> | -m <client> <to-server> <from-server> | -h\n";

	fprintf(stderr, oosage, name, name);
}

u_int* migrate(const char *client, const char *to_server)
{
	static u_int er;
	static u_int *erp;
	struct in_addr caddr, saddr;
	struct s_trans_migrate_cmd cmd;
	
	er = S_TRANS_STATUS_FAILED;
	erp = &er;

	/* in network byte order .. */
	printf("%s: trying to migrate client %s to server %s\n", __FUNCTION__, client, to_server);
	if (in_addr_for_host(&caddr, client))
		goto out;
	if (in_addr_for_host(&saddr, to_server))
		goto out;
	cmd.sm_client_ip = caddr.s_addr;
	cmd.sm_server_ip = saddr.s_addr;

	erp = s_trans_migrate_client_1(&cmd, clnt);
out:
	return erp;
}

int main(int argc, char *argv[])
{
	char *flag, *arg, *arg2, *server = NULL;
	u_int *err;
	struct in_addr addr;

	if (argc == 3) {
		server = argv[2];
	} else if (argc == 4) {
		arg = argv[2];
		server = argv[3];
	} else if (argc == 5) {
		/* temporary: just for the stupid first try */
		arg = argv[2];
		arg2 = argv[3];
		server = argv[4];
	} else {
		usage();
		exit(1);
	}
	flag = argv[1];

	clnt = clnt_create(server, S_TRANS_PROG, S_TRANS_VERS, "tcp");
	if (clnt == NULL) {
		clnt_pcreateerror(server);
		exit (1);
	}

	if (is_flag("-t") || is_flag("--test")) {
		err = s_trans_void_1((void *)NULL, clnt);
		if (*err != S_TRANS_STATUS_SUCCESS)
			fprintf(stderr, "Error during test RPC to server %s.\n", server);
		else
			printf("RPC test to server %s was successful.\n", server);
	} else if (is_flag("-l")) {
		struct s_trans_simple_cmd scmd;

		scmd.sm_type = S_TRANS_LOAD_REBOOT_STATE;
		err = s_trans_load_reboot_state_1(&scmd, clnt);
		if (*err != S_TRANS_STATUS_SUCCESS)
			fprintf(stderr, "Error while attempting to load reboot state on server %s.\n", server);
		else
			printf("Recovery state loaded on server %s.\n", server);
	} else if (is_flag("-a")) {
		/* XXX: we need to check the error code .. */
		struct s_trans_add_cmd acmd;
		struct s_trans_str sstr;

		if (in_addr_for_host(&addr, arg2))
			fprintf(stderr, "Error while attempting to resolve IP for client '%s'.\n", arg2);		

		acmd.sm_type = S_TRANS_ADD_STATE;
		acmd.sm_data.sm_data_len = 1;
		acmd.sm_data.sm_data_val = &sstr;
		sstr.text = arg;
		acmd.sm_client_ip = addr.s_addr;
		acmd.sm_client_boottime = 1111111111;
		printf("WARNING: sending client boottime as '1111111111', which is a long time ago.\n");
		
		err = s_trans_add_state_1(&acmd, clnt);
		if (*err != S_TRANS_STATUS_SUCCESS)
			fprintf(stderr, "Error while attempting to add client-recovery-dir '%s' on server %s.\n", arg, server);
		else
			printf("Added client-recovery-dir '%s' on server %s.\n", arg, server);
	} else if (is_flag("-r")) {
		/* XXX: we need to check the error code .. */
		struct s_trans_remove_cmd rcmd;
		struct s_trans_str sstr;

		if (in_addr_for_host(&addr, arg2))
			fprintf(stderr, "Error while attempting to resolve IP for client '%s'.\n", arg2);		

		rcmd.sm_type = S_TRANS_REMOVE_STATE;
		rcmd.sm_data.sm_data_len = 1;
		rcmd.sm_data.sm_data_val = &sstr;
		sstr.text = arg;

		err = s_trans_remove_state_1(&rcmd, clnt);
		if (*err != S_TRANS_STATUS_SUCCESS)
			fprintf(stderr, "Error while attempting to remove client-recovery-dir '%s' on server %s.\n", arg, server);
		else
			printf("Removed client-recovery-dir '%s' on server %s.\n", arg, server);
	} else if (is_flag("-m")) {
		err = migrate(arg, arg2);
	} else if (is_flag("-g")) {
		struct s_trans_simple_cmd gcmd;

		gcmd.sm_type = S_TRANS_ENTER_GRACE;
		err = s_trans_enter_grace_1(&gcmd, clnt);
		if (*err != S_TRANS_STATUS_SUCCESS)
			fprintf(stderr, "Error while attempting to have server %s enter grace.\n", server);
		else
			printf("Server %s entered grace.\n", server);
	} else {
		fprintf(stderr, "Unrecognized flag: %s\n", flag);
		exit(1);
	}
	
	exit(*err);
}
