#include <dirent.h>
#include "statetransferd.h"

u_int *
s_trans_void_1_svc(void *cmd, struct svc_req *rqstp)
{
	static u_int result = S_TRANS_STATUS_SUCCESS;

	return &result;
}

u_int *
s_trans_load_reboot_state_1_svc(struct s_trans_simple_cmd *cmd, struct svc_req *rqstp)
{
	static u_int result;
	static struct s_trans_msg stm;
	static char *wbuf = NULL, *host;
	static int wbuflen, wrote;

	result = S_TRANS_STATUS_FAILED;
	
	/* make a downcall to pass along the request .. */
	stm.sm_status = S_TRANS_STATUS_SUCCESS;
	stm.sm_type = S_TRANS_LOAD_REBOOT_STATE;
	stm.sm_origin = S_TRANS_RPC;
	stm.sm_len = 0;

	/* do_state_trans() will allocate wbuf on LOAD. */
	/* if we have an error, don't make a downcall, since knfsd didn't make an upcall. */
	if (do_state_trans(&stm, &wbuf, &wbuflen))
		goto out;
	if ((wrote = atomicio(write, s_trans_pipe_fd, wbuf, wbuflen)) != wbuflen) {
		syslog(LOG_INFO, "%s: write of %d bytes failed (wrote %d). %s\n", __FUNCTION__, wbuflen, wrote, strerror(errno));
		goto out;
	} else {
		result = S_TRANS_STATUS_SUCCESS;
	}

	host = inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr);
	syslog(LOG_INFO, "%s: relayed LOAD request from host '%s'.\n", __FUNCTION__, host);
out:
	if (wbuf)
		free(wbuf);
	return &result;
}

u_int *
s_trans_add_state_1_svc(struct s_trans_add_cmd *cmd, struct svc_req *rqstp)
{
	static u_int result, datalen;
	static char *host, *p;
	struct s_trans_msg *stmp = NULL;

	result = S_TRANS_STATUS_FAILED;

	/* should contain 1 client recovery dir */
	if (cmd->sm_data.sm_data_len != 1) {
		syslog(LOG_INFO, "%s: error -- didn't receive exactly 1 client recdir argument (%d).\n", __FUNCTION__, cmd->sm_data.sm_data_len);
		goto out;
	}
	
	/* asciified-hex client recdir, u32 client IP, u32 client boottime */
	datalen = HEXDIR_LEN_PADDED + (2 * sizeof(uint32_t));
	p = calloc(1, sizeof(*stmp) + datalen);
	if (p == NULL)
		goto out;
	stmp = (struct s_trans_msg *)p;

	/* we still fill this out for the daemon, despite that we don't make a downcall */
	stmp->sm_status = S_TRANS_STATUS_SUCCESS;
	stmp->sm_type = S_TRANS_ADD_STATE;
	stmp->sm_origin = S_TRANS_RPC;
	stmp->sm_len = datalen;

	p += sizeof(*stmp);
	memcpy(p, cmd->sm_data.sm_data_val[0].text, HEXDIR_LEN);
	p += HEXDIR_LEN_PADDED;
	memcpy(p, &cmd->sm_client_ip, sizeof(uint32_t));
	p += sizeof(uint32_t);
	memcpy(p, &cmd->sm_client_boottime, sizeof(uint32_t));
		
	if (do_state_trans(stmp, NULL, 0)) {
		syslog(LOG_INFO, "%s: failed ADD request.\n", __FUNCTION__);
		goto out;
	}

	result = S_TRANS_STATUS_SUCCESS;
	host = inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr);
	syslog(LOG_INFO, "%s: processed ADD request from host '%s'.\n", __FUNCTION__, host);
out:
	if (stmp)
		free(stmp);
	return &result;
}

u_int *
s_trans_remove_state_1_svc(struct s_trans_remove_cmd *cmd, struct svc_req *rqstp)
{
	static u_int result, datalen;
	static char *host, *p;
	struct s_trans_msg *stmp = NULL;

	result = S_TRANS_STATUS_FAILED;

	/* should contain 1 client recovery dir */
	if (cmd->sm_data.sm_data_len != 1) {
		syslog(LOG_INFO, "%s: error -- did not receive exactly 1 client-recovery-dir argument (%d).\n", __FUNCTION__, cmd->sm_data.sm_data_len);
		goto out;
	}

	/* asciified-hex client recdir */
	datalen = HEXDIR_LEN_PADDED;
	p = calloc(1, sizeof(*stmp) + datalen);
	if (p == NULL)
		goto out;
	stmp = (struct s_trans_msg *)p;

	/* we still fill this out for the daemon, even though we don't make a downcall */
	stmp->sm_status = S_TRANS_STATUS_SUCCESS;
	stmp->sm_type = S_TRANS_REMOVE_STATE;
	stmp->sm_origin = S_TRANS_RPC;
	stmp->sm_len = datalen;

	p += sizeof(*stmp);
	memcpy(p, cmd->sm_data.sm_data_val[0].text, HEXDIR_LEN);

	if (do_state_trans(stmp, NULL, 0)) {
		syslog(LOG_INFO, "%s: failed REMOVE request.\n", __FUNCTION__);
		goto out;
	}

	result = S_TRANS_STATUS_SUCCESS;
	host = inet_ntoa(rqstp->rq_xprt->xp_raddr.sin_addr);
	syslog(LOG_INFO, "%s: processed REMOVE request from host '%s'.\n", __FUNCTION__, host);
out:
	if (stmp)
		free(stmp);
	return &result;
}

/* addresses are in network byte order.
 * like most of this code, assumes we're already in 'nfsd_recovery_dir'.
 * XXX: what a long function
 */
u_int *
s_trans_migrate_client_1_svc(struct s_trans_migrate_cmd *cmd, struct svc_req *rqstp)
{
	static u_int result, buflen;
	uint *err;
	int i, nents, wrote, did_add = 0;
	char *c, *subdir, *server = NULL;
	char ipfile[CLT_IPFILE_PATHLEN], ipstr[MAX_IP4_STRLEN];
	FILE *fp;
	struct dirent **ents;
	struct in_addr iaddr;
	CLIENT *s_trans_clnt = NULL;
	struct s_trans_msg *stmp;
	struct s_trans_simple_cmd lcmd;
	struct s_trans_simple_cmd gcmd;
	struct s_trans_add_cmd acmd;
	struct s_trans_str sstr;

	result = S_TRANS_STATUS_FAILED;

	/* scan the recovery dir and find the clientids we need to migrate */
	nents = scandir(nfsd_recovery_dir, &ents, NULL, alphasort);
	if (nents < 0) {
		syslog(LOG_INFO, "%s: unable to scan recovery dir '%s'.\n", __FUNCTION__, nfsd_recovery_dir);
		goto out;
	} else if (nents == 0) {
		syslog(LOG_INFO, "%s: no entries in recovery dir '%s'.\n", __FUNCTION__, nfsd_recovery_dir);
		goto out;
	}

	for (i = 0; i < nents; i++) {
		subdir = ents[i]->d_name;
		if (!strcmp(subdir, ".") || !strcmp(subdir, ".."))
			continue;
		else if (strlen(subdir) != HEXDIR_LEN) {
			//syslog(LOG_INFO, "%s: note -- recovery dir '%s' is not HEXDIR_LEN.\n", __FUNCTION__, subdir);
			continue;
		}

		/* open the ipfile and read the clientIP */
		if (sprintf(ipfile, "%s/%s", subdir, CLT_IPFILE) < 0) {
			syslog(LOG_INFO, "%s: failed building path.\n", __FUNCTION__);
			continue;
		}
		if ((fp = fopen(ipfile, "r")) == NULL) {
			syslog(LOG_INFO, "%s: failed opening ipfile '%s'.\n", __FUNCTION__, ipfile);
			continue;
		}
		if (fgets(ipstr, MAX_IP4_STRLEN, fp) == NULL) {
			syslog(LOG_INFO, "%s: couldn't read clientip from '%s'.\n", __FUNCTION__, ipfile);
			fclose(fp);
			continue;
		}

		/* try matching the clientIP to the requested migratee */
		if (inet_aton(ipstr, &iaddr) == 0) {
			syslog(LOG_INFO, "%s: unable to convert IP '%s' to an in_addr.\n", __FUNCTION__, ipstr);
			continue;
		}
		if (iaddr.s_addr != cmd->sm_client_ip) {
			//syslog(LOG_INFO, "%s: '%s' converted to '%x', but it didn't match requested '%x'\n", __FUNCTION__, ipstr, iaddr.s_addr, cmd->sm_client_ip);
			continue;
		}
	
		/* make sure we have an RPC client .. */
		if (s_trans_clnt == NULL) {
			server = dotted_quad_for_long(cmd->sm_server_ip);
			s_trans_clnt = clnt_create(server, S_TRANS_PROG, S_TRANS_VERS, "tcp");
			if (s_trans_clnt == NULL) {
				clnt_pcreateerror(server);
				syslog(LOG_INFO, "%s: unable to make RPC client; failing..\n", __FUNCTION__);
				goto out;
			}
		}
			
		/* send out ADD_STATE for this clientIP */
		acmd.sm_type = S_TRANS_ADD_STATE;
		acmd.sm_data.sm_data_len = 1;
		acmd.sm_data.sm_data_val = &sstr;
		sstr.text = subdir;
		acmd.sm_client_ip = cmd->sm_client_ip;
		acmd.sm_client_boottime = extract_client_boottime(subdir);
		
		syslog(LOG_INFO, "%s: attempting ADD_STATE RPC to server '%s' (dir %s  bootttime %d).\n", __FUNCTION__, server, subdir, acmd.sm_client_boottime);
		err = s_trans_add_state_1(&acmd, s_trans_clnt);
		if (*err != S_TRANS_STATUS_SUCCESS) {
			syslog(LOG_INFO, "%s: ADD_STATE RPC to server '%s' (re: client '%s') failed.\n", __FUNCTION__, server, subdir);
		} else {
			syslog(LOG_INFO, "%s: ADD_STATE RPC to server '%s' (re: client '%s') succeeded.\n", __FUNCTION__, server, subdir);
			did_add = 1;
		}
	} /* foreach candidate recovery dir .. */

	if (!did_add) {
		syslog(LOG_INFO, "%s: unable to add state/migrate -- apparently no clients associated with client ip 0x%.8x\n", __FUNCTION__, cmd->sm_client_ip);
		goto out;
	}


	/* XXX: note that this ENTER_GRACE is basically broken -- every server in
	 *   the pool -should- enter grace together, but we don't have a list of servers in
	 *   the pool.  furthermore, if that list ends up living in-kernel, it might be
	 *   proper to issue the following client RPC from within knfsd itself.
	 *
	 *   of course, once we get real state transfer stuff working, no grace will be nec.
	 */

	/* send out ENTER_GRACE .. */
	gcmd.sm_type = S_TRANS_ENTER_GRACE;
	err = s_trans_enter_grace_1(&gcmd, s_trans_clnt);
	if (*err != S_TRANS_STATUS_SUCCESS) {
		syslog(LOG_INFO, "%s: ENTER_GRACE RPC to server '%s' failed.\n", __FUNCTION__, server);
		goto out;
	} else
		syslog(LOG_INFO, "%s: ENTER_GRACE RPC to server '%s' succeeded.\n", __FUNCTION__, server);

	/* tell our own knfsd to enter grace .. */
	err = s_trans_enter_grace_1_svc(&gcmd, rqstp);
	if (*err != S_TRANS_STATUS_SUCCESS) {
		syslog(LOG_INFO, "%s: ENTER_GRACE msg to knfsd failed.\n", __FUNCTION__);
		goto out;
	} else
		syslog(LOG_INFO, "%s: ENTER_GRACE msg to knfsd succeeded.\n", __FUNCTION__);

	/* send out LOAD_STATE .. */
	lcmd.sm_type = S_TRANS_LOAD_REBOOT_STATE;
	err = s_trans_load_reboot_state_1(&lcmd, s_trans_clnt);
	if (*err != S_TRANS_STATUS_SUCCESS) {
		syslog(LOG_INFO, "%s: LOAD_STATE RPC to server '%s' failed.\n", __FUNCTION__, server);
		goto out;
	} else
		syslog(LOG_INFO, "%s: LOAD_STATE RPC to server '%s' succeeded.\n", __FUNCTION__, server);
	
	/* tell our knfsd to set up the migration order .. */
	buflen = sizeof(*stmp) + (2 * sizeof(struct in_addr));
	c = calloc(1, buflen);
	stmp = (struct s_trans_msg *)c;
	stmp->sm_status = S_TRANS_STATUS_SUCCESS;
	stmp->sm_type = S_TRANS_MIGRATE_CLIENT;
	stmp->sm_origin = S_TRANS_RPC;
	stmp->sm_len = (2 * sizeof(struct in_addr));

	c += sizeof(*stmp);
	memcpy(c, &cmd->sm_client_ip, sizeof(struct in_addr));
	c += sizeof(struct in_addr);
	memcpy(c, &cmd->sm_server_ip, sizeof(struct in_addr));

	if ((wrote = atomicio(write, s_trans_pipe_fd, stmp, buflen)) != buflen) {
		syslog(LOG_INFO, "%s: 'migrate' knfsd msg failed.\n", __FUNCTION__);
		goto out;
	} else
		syslog(LOG_INFO, "%s: 'migrate' knfsd msg succeeded.\n", __FUNCTION__);

	result = S_TRANS_STATUS_SUCCESS;
out:
	if (server)
		free(server);
	return &result;
}

u_int *
s_trans_enter_grace_1_svc(struct s_trans_simple_cmd *cmd, struct svc_req *rqstp)
{
	static u_int result;
	static struct s_trans_msg stm;
	static int wrote;

	stm.sm_status = S_TRANS_STATUS_SUCCESS;
	stm.sm_type = S_TRANS_ENTER_GRACE;
	stm.sm_origin = S_TRANS_RPC;
	stm.sm_len = 0;

	if ((wrote = atomicio(write, s_trans_pipe_fd, ((void *)&stm), sizeof(stm))) != sizeof(stm)) {
		syslog(LOG_INFO, "%s: (grace) write of %lu bytes failed (wrote %d). %s\n", __FUNCTION__, sizeof(stm), wrote, strerror(errno));
		result = S_TRANS_STATUS_FAILED;
	} else {
		syslog(LOG_INFO, "%s: (grace) downcall to knfsd seemed good.\n", __FUNCTION__);
		result = S_TRANS_STATUS_SUCCESS;
	}

	return &result;
}

u_int *
s_trans_exit_grace_1_svc(struct s_trans_simple_cmd *cmd, struct svc_req *rqstp)
{
	static u_int  result;

	result = S_TRANS_STATUS_SUCCESS;
	syslog(LOG_INFO, "%s: EXIT_GRACE command isn't implemented, as there isn't a \"known-servers\" list.\n", __FUNCTION__);

	return &result;
}

