#ifndef STATETRANSFERD_H
#define STATETRANSFERD_H 1

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>

#include "statetransfer.h"


/*------------- start common with kernel -----------
 *  * see include/linux/nfsd/statetrans.h
 */

/* Message types */
#define S_TRANS_VOID                    1
#define S_TRANS_LOAD_REBOOT_STATE       2 /* takes list of clientids */
#define S_TRANS_ADD_STATE               3
#define S_TRANS_REMOVE_STATE            4
#define S_TRANS_ENTER_GRACE             5
#define S_TRANS_EXIT_GRACE              6
#define S_TRANS_MIGRATE_CLIENT		7

/* Status */
#define S_TRANS_STATUS_INVALIDMSG       0x01
#define S_TRANS_STATUS_AGAIN            0x02
#define S_TRANS_STATUS_LOOKUPFAIL       0x04
#define S_TRANS_STATUS_SUCCESS          0x08
#define S_TRANS_STATUS_FAILED           0x10
#define S_TRANS_STATUS_BADSERVERIP      0x20
#define S_TRANS_STATUS_BADCLIENTIP      0x40
#define S_TRANS_STATUS_BADCLIENTID      0x80

/* Originator of transfer request */
#define S_TRANS_KNFSD			1
#define S_TRANS_RPC			2

/* rpc pipe - MUST be co-ordinated with kernel */
#define S_TRANS_PIPE "statetrans"
#define ST_DELAY        2*HZ  /* Time to wait for open on pipe */
#define ST_MAX          4096  /* Maximum message length */

/*
 * The message header.
 */
struct s_trans_msg {
	uint32_t     sm_status;
	uint32_t     sm_type;
	uint32_t     sm_origin; /* S_TRANS_KNFSD or S_TRANS_RPC */
	uint32_t     sm_len;    /* length of following data */
};

/*------------- end common with kernel -----------*/


/* NB: Here, HEXDIR_LEN is 32 -- the number of characters in the name.
 *     In the kernel, HEXDIR_LEN is 33 -- as above, but plus a null-term.
 *     We need to send the null, but it also has to be word-aligned for
 *     XDR encoding, so we pad it out to 36.
 */
#define HEXDIR_LEN		32 /* hex version of 16 byte md5 of cl_name */
#define HEXDIR_LEN_PADDED	36 /* as above, null-termed and word-aligned */
#define XDR_QUADLEN(l)		(((l) + 3) >> 2)

#define HEXIP_LEN 		8  /* 8 hex chars for a u32 IP */
#define MAX_IP4_STRLEN		16 /* dotted-quad and null-term */

#define SVR_CUR_TIMEFILE	"server_boottime"
#define SVR_PREV_TIMEFILE	"prev_server_boottime"
#define CLT_TIMEFILE		"client_boottime"
#define CLT_TIMEFILE_PATHLEN	(HEXDIR_LEN + 1 + strlen(CLT_TIMEFILE) + 1)
#define CLT_IPFILE		"client_ip"
#define CLT_IPFILE_PATHLEN	(HEXDIR_LEN + 1 + strlen(CLT_IPFILE) + 1)
#define EPOCH_STRLEN		11 /* 10 ASCII digits and null-term */


//#define NFSD_PIPEFS_DIR "/tmp/var/lib/nfs/rpc_pipefs/nfsd"

#ifndef NFSD_PIPEFS_DIR  
#define NFSD_PIPEFS_DIR		"/var/lib/nfs/rpc_pipefs/nfsd"
#endif

#ifndef S_TRANS_PIPE_PATH 
#define S_TRANS_PIPE_PATH	NFSD_PIPEFS_DIR "/" S_TRANS_PIPE
#endif

#ifndef NFSD_RECOVERY_DIR
#define NFSD_RECOVERY_DIR	"/var/lib/nfs/v4recovery"
#endif


/* XXX: perhaps move "start common with kernel" stuff into nfsd_statetransfer.h
 *           and the "new header" into statetransferd.h
 */

/* statetransfer_svc.c */
extern void s_trans_prog_1(struct svc_req *, register SVCXPRT *);

/* atomicio.c */
extern ssize_t atomicio(ssize_t (*)(), int, void *, size_t);

/* statetransferd.c */
extern int do_state_trans(struct s_trans_msg *, char **, int *);

extern char *rpc_pipefs_dir;
extern char *nfsd_pipefs_dir;
extern char *s_trans_pipe_path;
extern char *nfsd_recovery_dir;
extern int rpc_pipefs_fd;
extern int nfsd_pipefs_fd;
extern int s_trans_pipe_fd;

/* statetransfer_clnt.c */
extern u_int * s_trans_add_state_1(struct s_trans_add_cmd *, CLIENT *);

/* misc.c */
extern int in_addr_for_host(struct in_addr *, const char *);
extern char * dotted_quad_for_long(unsigned long);
extern char * dotted_quad_for_host(const char *);
extern time_t extract_boottime(const char *);
extern time_t extract_client_boottime(const char *);
extern time_t set_server_boottime(time_t);


#endif
